/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * Configuration management.
 * @author Christian Grothoff
 * @file common/configuration.c
 **/

#include "configuration.h"
#include "config.h"
#include "util/returnvalues.h"
#include "util/storage.h"
#include "util/ipcheck.h"
#include "util/logging.h"

#include <netdb.h>
#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>



#define NETWORK_SECTION       "NETWORK"
#define NETWORK_IP            "IP"
#define NETWORK_INTERFACES    "INTERFACES"
#define NETWORK_PORT          "PORT"
#define NETWORK_DOWNBPS_TOTAL "MAXNETDOWNBPSTOTAL"
#define NETWORK_UPBPS_TOTAL   "MAXNETUPBPSTOTAL"
#define NETWORK_MAX_LOAD      "MAXNETLOAD"
#define NETWORK_NOHELO        "NOHELO"
#define NETWORK_BLACKLIST     "BLACKLIST"
#define NETWORK_TRUSTED       "TRUSTED"
#define NETWORK_HOSTLISTURL   "HOSTLISTURL"

#define FILESHARING_SECTION   "FILESHARING"
#define FILESHARING_MIGRATION "ACTIVEMIGRATION"
#define FILESHARING_TIMEOUT   "SEARCHTIMEOUT"
#define FILESHARING_EXTRACTOR "EXTRACTORS"

#define FILES_SECTION         "FILES"
#define FILES_QUOTA           "DISKQUOTA"
#define FILES_HOSTKEY         "HOSTKEY"
#define FILES_DATABASE        "DATABASE"
#define FILES_AGEFILE         "AGEFILE"
#define FILES_SHARED_LIST     "FILELIST"
#define FILES_CONTENT_DIR     "CONTENTDIR"
#define FILES_CREDIT_DIR      "CREDITDIR"
#define FILES_HOSTS_DIR       "HOSTS"

#define GNUNETD_SECTION       "GNUNETD"
#define GNUNETD_MAX_NODES     "MAXNODES"
#define GNUNETD_HELO_EXPIRES  "HELOEXPIRES"
#define GNUNETD_MAX_LOAD      "MAXLOAD"

#define GNUNETD_LOGLEVEL       "LOGLEVEL"
#define GNUNETD_LOGFILE        "LOGFILE"
#define GNUNETD_PID            "PIDFILE"

/**
 * PRIVATE: gnet configuration
 **/
LHASH * config_;

static FileName sharedFileList_;
static FileName creditDirectory_;
static FileName contentDirectory_;
static FileName hostDirectory_;
static FileName ageFile_;
static FileName privHostKey_;
static FileName databaseFile_;
static FileName logFile_;
static FileName pidFile_;
static FileName configuration_filename;
static CIDRNetwork * filteredNetworks_;
static CIDRNetwork * trustedNetworks_;
static Mutex configLock;

static FileName getFileName(char * section,
			    char * value,
			    char * errMsg) {
  FileName fn;
  fn = CONF_get_string(config_,
		       section,
		       value);
  if (fn == NULL) {
    errexit(errMsg,
	    section,
	    value);
  }
  return expandFileName(fn);   
}


static void readConfig(FileName fn) {
  long errorLine; /* in which line did an error occur */
  char * ch;
  
  if (!assertIsFile(fn))
    errexit("Configuration file %s is not a file (or does not exist)\n", 
	    fn);
  INFO("Loading configuration from %s.\n", fn);
  config_ = CONF_load(NULL, fn, &errorLine);
  if (config_ == NULL) {
    ERR_load_crypto_strings();
    FAILURE("Unable to load configuration from %s. Error at line %ld.\n",
	    fn,
	    errorLine);
    ERR_print_errors_fp(stderr);
    errexit("Bad configuration file %s.\n", 
	    fn);
  }
  privHostKey_ = 
    getFileName(FILES_SECTION,
		FILES_HOSTKEY,
		"Configuration file must specify filename for hostkey in section %s under %s.\n");
  sharedFileList_ = 
    getFileName(FILES_SECTION,
		FILES_SHARED_LIST,
		"Configuration file must specify filename for the list of on-demand shared files in section %s under %s.\n");
  creditDirectory_ =
    getFileName(FILES_SECTION,
		FILES_CREDIT_DIR,
		"Configuration file must specify directory for credits in section %s under %s.\n");
  contentDirectory_ =
    getFileName(FILES_SECTION,
		FILES_CONTENT_DIR,
		"Configuration file must specify directory for content in section %s under %s.\n");
  hostDirectory_ =
    getFileName(FILES_SECTION,
		FILES_HOSTS_DIR,
		"Configuration file must specify directory for network identities in section %s under %s.\n");
  databaseFile_ =
    getFileName(FILES_SECTION,
		FILES_DATABASE,
		"Configuration file must specify filename for the database in section %s under %s.\n");
  ageFile_ =
    getFileName(FILES_SECTION,
		FILES_AGEFILE,
		"Configuration file must specify filename for the database in section %s under %s.\n");
  logFile_ =
    getFileName(GNUNETD_SECTION,
		GNUNETD_LOGFILE,
		"Configuration file must specify filename for the logs in section %s under %s.\n");
  pidFile_ = 
    getFileName(GNUNETD_SECTION,
		GNUNETD_PID,
		"Configuration file must specify a filename where gnunetd writes its process id (PID) in section %s under %s\n");
  /* create directories */
  mkdirp(hostDirectory_);
  mkdirp(creditDirectory_);
  mkdirp(contentDirectory_);
  ch = CONF_get_string(config_,
		       NETWORK_SECTION,
		       NETWORK_BLACKLIST);
  if (ch == NULL)
    ch = "";
  filteredNetworks_ = parseRoutes(ch);
  if (filteredNetworks_ == NULL) 
    errexit("Malformed entry in the configuration in section %s under %s: %s\n",
	    NETWORK_SECTION,NETWORK_BLACKLIST,ch);  
  ch = CONF_get_string(config_,
		       NETWORK_SECTION,
		       NETWORK_TRUSTED);
  if (ch == NULL)
    ch = "127.0.0.0/8;"; /* by default, trust localhost only */
  trustedNetworks_ = parseRoutes(ch);
  if (trustedNetworks_ == NULL) 
    errexit("Malformed entry in the configuration in section %s under %s: %s\n",
	    NETWORK_SECTION,NETWORK_TRUSTED,ch);  
}

static void freeConfig() {
  INFO("Freeing configuration.\n");
  CONF_free(config_);
  xfree(privHostKey_, "freeConfig: privHostKey");
  xfree(sharedFileList_, "freeConfig: sharedFileList");
  xfree(creditDirectory_, "freeConfig: creditDirectory");
  xfree(contentDirectory_, "freeConfig: contentDirectory");
  xfree(hostDirectory_, "freeConfig: hostDirectory ");
  xfree(databaseFile_, "freeConfig: databaseFile");
  xfree(ageFile_, "freeConfig: ageFile");
  xfree(filteredNetworks_, "freeConfig: filteredNetworks");
  xfree(trustedNetworks_, "freeConfig: trustedNetworks");
}

void initConfiguration(FileName filename) {
  configuration_filename = expandFileName(filename);

  readConfig(configuration_filename);
  create_mutex(&configLock);
}

void updateConfiguration() {
  long errorLine; /* in which line did an error occur */

  MUTEX_LOCK(&configLock);
  CONF_free(config_);
  if (!assertIsFile(configuration_filename))
    errexit("Configuration file %s is not a file", 
	    configuration_filename);
  INFO("Loading configuration from %s.\n", configuration_filename);
  config_ = CONF_load(NULL, configuration_filename, &errorLine);
  if (config_ == NULL) {
    ERR_load_crypto_strings();
    FAILURE("Unable to load configuration from %s. Error at line %ld.\n",
	    configuration_filename,
	    errorLine);
    ERR_print_errors_fp(stderr);
    errexit("Bad configuration file %s.", 
	    configuration_filename);
  }  
  MUTEX_UNLOCK(&configLock);
}

void doneConfiguration() {
  freeConfig();
  xfree(configuration_filename,
	"doneConfiguration: config filename");
  destroy_mutex(&configLock);
}


/* **************** network section ***************** */

/**
 * Defaults in case of invalid values in config file 
 **/
#define DEFAULT_MAXDOWNBPS 500000
#define DEFAULT_MAXUPBPS 500000

unsigned int getMaxDownstreamBPSTotal() {
  unsigned int ret;

  MUTEX_LOCK(&configLock);
  ret = (unsigned int) CONF_get_number(config_, 
					NETWORK_SECTION,
					NETWORK_DOWNBPS_TOTAL);
  MUTEX_UNLOCK(&configLock);
  if (ret <= 0)
    ret = DEFAULT_MAXDOWNBPS;
  return ret;
}

unsigned int getMaxUpstreamBPSTotal() {
  unsigned int ret;

  MUTEX_LOCK(&configLock);
  ret = (unsigned int) CONF_get_number(config_, 
				       NETWORK_SECTION,
				       NETWORK_UPBPS_TOTAL);
  MUTEX_UNLOCK(&configLock);
  if (ret <= 0)
    ret = DEFAULT_MAXUPBPS;
  return ret;
}

unsigned int getMaxNetLoad() {
  unsigned int ret;

  MUTEX_LOCK(&configLock);
  ret = (unsigned int) CONF_get_number(config_, 
				       NETWORK_SECTION,
				       NETWORK_MAX_LOAD);
  MUTEX_UNLOCK(&configLock);
  return ret;
}

IPAddress getIP() {
  IPAddress ip;
  MUTEX_LOCK(&configLock);  
  ip = CONF_get_string(config_,
		       NETWORK_SECTION,
		       NETWORK_IP);
  if (ip != NULL)
    ip = strdup(ip);
  MUTEX_UNLOCK(&configLock);
  return ip;
}

Interfaces getNetInterfaces() {
  Interfaces i;
  
  MUTEX_LOCK(&configLock);
  i = CONF_get_string(config_,
		      NETWORK_SECTION,
		      NETWORK_INTERFACES);
  if (i != NULL)
    i = strdup(i);
  MUTEX_UNLOCK(&configLock);
  return i;
}

unsigned short getGNUnetPort() {
  struct servent * pse;	/* pointer to service information entry	*/
  unsigned short port;

  MUTEX_LOCK(&configLock);
  port = (unsigned short) CONF_get_number(config_, 
				       NETWORK_SECTION,
				       NETWORK_PORT);
  if (port == 0) { /* try lookup in services */
    if ((pse = getservbyname("gnunet", "udp"))) 
      port = htons(pse->s_port);      
    else 
      errexit("Cannot determine port to bind to. Define in configuration file in section %s under %s or in /etc/services under UDP/gnunet.\n",
	      NETWORK_SECTION, NETWORK_PORT);
  }
  MUTEX_UNLOCK(&configLock);
  return port;
}

int isHELOEnabled() {
  unsigned char * option;
  int ret;

  MUTEX_LOCK(&configLock);
  option = CONF_get_string(config_,
			   NETWORK_SECTION,
			   NETWORK_INTERFACES);
  if (option == NULL)
    ret = YES; /* default: is gateway */
  else {
    if (strcmp(option,"YES"))
      ret = YES;
    else
      ret = NO;
  }
  MUTEX_UNLOCK(&configLock);
  return ret;
}

int isBlacklisted(struct in_addr ip) {
  int ret;

  MUTEX_LOCK(&configLock);
  ret = checkIPListed(filteredNetworks_,
		      ip);
  MUTEX_UNLOCK(&configLock);
  return ret;
}

int isWhitelisted(struct in_addr ip) {
  int ret;

  MUTEX_LOCK(&configLock);
  ret = checkIPListed(trustedNetworks_,
		      ip);
  MUTEX_UNLOCK(&configLock);
  return ret;
}

char * getHostlistURL() {
  char * i;

  MUTEX_LOCK(&configLock);
  i = CONF_get_string(config_,
		      NETWORK_SECTION,
		      NETWORK_HOSTLISTURL);
  if (i != NULL)
    i = strdup(i);
  MUTEX_UNLOCK(&configLock);
  return i;
}

/* ********************* Filesharing options ********************* */

unsigned int getSearchTimeout() {
  unsigned int ret;
  
  MUTEX_LOCK(&configLock);
  ret = (unsigned int) CONF_get_number(config_, 
					FILESHARING_SECTION,
					FILESHARING_TIMEOUT);
  MUTEX_UNLOCK(&configLock);
  return ret;
}

int activeMigration() {
  char *  f;
  int ret;

  MUTEX_LOCK(&configLock);
  f = CONF_get_string(config_,
		      FILESHARING_SECTION,
		      FILESHARING_MIGRATION);
  if (f == NULL)
    ret = NO; /* default: NO */
  else {
    if (0 == strcmp(f,"YES") )
      ret = YES;
    else
      ret = NO;
  }
  MUTEX_UNLOCK(&configLock);
  return ret;
}

char * getExtractorConfig() {
  char * cfg;

  MUTEX_LOCK(&configLock);
  cfg = CONF_get_string(config_,
			FILESHARING_SECTION,
			FILESHARING_EXTRACTOR);
  if (cfg != NULL)
    cfg = strdup(cfg);
  MUTEX_UNLOCK(&configLock);
  return cfg;
}
/* ********************* File options ********************* */

unsigned int getDiskQuota() {
  unsigned int ret;

  MUTEX_LOCK(&configLock);
  ret = (unsigned int) CONF_get_number(config_, 
				       FILES_SECTION,
				       FILES_QUOTA);
  MUTEX_UNLOCK(&configLock);
  return ret;
}

FileName getHostKeyFilePriv() {
  return privHostKey_;
}

FileName getAgeFile() {
  return ageFile_;
}

FileName getDatabaseFile() {
  return databaseFile_;
}

FileName getSharedFileList() {
  return sharedFileList_;
}

FileName getCreditDirectory() {
  return creditDirectory_;
}

FileName getContentDirectory() {
  return contentDirectory_;
}

FileName getHostsDirectory() {
  return hostDirectory_;
}

FileName getPIDFile() {
  return pidFile_;
}

/* ************* server options ***************** */

unsigned int getMaxLoad() {
  unsigned int ret;

  MUTEX_LOCK(&configLock);
  ret = (unsigned int) CONF_get_number(config_,
				       GNUNETD_SECTION,
				       GNUNETD_MAX_LOAD);
  MUTEX_UNLOCK(&configLock);
  return ret;
}

unsigned int getMaxNodes() {
  unsigned int ret;

  MUTEX_LOCK(&configLock);
  ret = (unsigned int) CONF_get_number(config_, 
					GNUNETD_SECTION,
					GNUNETD_MAX_NODES);
  MUTEX_UNLOCK(&configLock);
  return ret;
}

unsigned int getLogLevel() {
  unsigned int ret;

  MUTEX_LOCK(&configLock);
  ret = (unsigned int) CONF_get_number(config_, 
				       GNUNETD_SECTION,
				       GNUNETD_LOGLEVEL);
  MUTEX_UNLOCK(&configLock);
  return ret;
}

FileName getLogFilename() {
  return logFile_;
}

/**
 * How many hours is a HELO message
 * valid?
 **/
unsigned int getHeloExpiration() {
  unsigned int exp;

  MUTEX_LOCK(&configLock);
  exp = (unsigned int) CONF_get_number(config_, 
				       GNUNETD_SECTION,
				       GNUNETD_HELO_EXPIRES);
  if (exp < 1) 
    exp = 1;
  MUTEX_UNLOCK(&configLock);
  return exp;
}




/* end of configuration.c */
