/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Little tool to do consistency check.
 * @file server/check_database.c
 * @author Christian Grothoff
 **/

#include "config.h"
#include "lookup.h"

#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>
#include "util/getopt.h"
#include "lookup.h"
#include "tcp_return.h"
#include "textui/insertfile.h"
#include "server/tcpserver.h"
#include "util/tcpio.h"
#include "util/contentdatabase.h"
#include "util/xmalloc.h"

extern char * indexed_files;

/* number of files that are directly shared */
extern unsigned int indexed_files_size;

/* configuration: do we fix problems? */
static int do_fix = YES;

/* tcp server result: were all the requests satisfied? */
static int tcp_verifies;

static int progressDot = 0;

static int be_verbose = NO;

static int be_quiet = NO;

/**
 * Check that the content at the given offset/file
 * has the given double hash.
 **/
static int checkHashMatch(int fileNameIndex,
			  size_t offset,
			  HashCode160 * doubleHash) {
  CONTENT_Block result;
  char * fn;
  HashCode160 hc;
  HashCode160 dhc;
  int fileHandle;
  size_t blen;
  /* check that the specified file at the specified
     offset actually contains the content that we
     are looking for */

  if ( (fileNameIndex < 1) ||
       (fileNameIndex > indexed_files_size) )
    return SYSERR;
  fn = indexed_files[fileNameIndex-1];
  fileHandle = open(fn, O_EXCL, S_IRUSR);
  if (fileHandle == -1) {
    WARNING("Could not open file %s (%d). ",
	  fn, fileNameIndex);    
    return SYSERR;
  }
  lseek(fileHandle, 
	offset, SEEK_SET);
  blen = read(fileHandle, 
	      &result,
	      sizeof(CONTENT_Block));
  close(fileHandle);
  hash(&result, 
       blen, 
       &hc);
  hash(&hc,
       sizeof(HashCode160),
       &dhc);
  if (0 != memcmp(&dhc,
		  doubleHash,
		  sizeof(HashCode160))) {
    return SYSERR;
  } else
    return OK;
}

/**
 * Check that the triple-hash 2 double-hash database
 * (lookup.c) has a mapping for this hashcode and
 * that the file-indices map it to the content database!
 **/
static void indexContent(HashCode160 * doubleHash) {
  ContentIndex entry;
  ContentIndex res;
  HashCode160 triple;
  HexName hn;

  hash2hex(doubleHash,
	   &hn);
  if (be_verbose == YES)
    printf("* %s (cdb)\n",
	  (char*)&hn);
  else {
    progressDot++;
    if ((progressDot & 255) == 0)
      if (be_quiet == NO) {
	printf(".");
	fflush(stdout);
      }
  }
  hash(doubleHash,
       sizeof(HashCode160),
       &triple);
  memcpy(&entry.doubleHash,
	 doubleHash,
	 sizeof(HashCode160));
  entry.importance = htonl(LOCAL_INDEXED_CONTENT_PRIO);
  entry.fileNameIndex = htonl(-1); /* directly available in data/content */
  entry.fileOffset = htonl(-1); 
  if (NO == findEntry(&triple,
		      &res)) {
    HexName hn;
    hash2hex(&entry.doubleHash,
	     &hn);
    if (be_quiet == NO)
      printf("Content %s not indexed in triple-hash database. ",
	     (char*) &hn);
    if (do_fix == YES) {	   
      if (SYSERR == addEntry(&entry)) {
	printf("addEntry returned SYSERR, can not fix!\n");
      } else {
	if (NO == findEntry(&triple, 
			    &res) )
	  printf("addEntry or findEntry had an internal error!\n");
	else
	  if (be_quiet == NO)
	    printf("Fixed.\n");
      }
    } else
      if (be_quiet == NO)
	printf("\n");
  } else {
    entry.importance = res.importance;
    if (0 != memcmp(&res, 
		    &entry,
		    sizeof(ContentIndex))) {
      if (SYSERR == checkHashMatch(ntohl(res.fileNameIndex),
				   ntohl(res.fileOffset),
				   &entry.doubleHash)) {
	/*	printContentIndex(&res);
		printContentIndex(&entry);*/
	if (be_quiet == NO)
	  printf("Bad value (I) stored in triple-hash database (could be ok) ");
	if (do_fix == YES) {
	  addEntry(&entry);
	  if (be_quiet == NO)
	    printf("Fixed.");
	} else
	  if (be_quiet == NO)
	    printf("\n");
      }
    }
  }
}

/**
 * Check that for each entry in the contentdatabase
 * there is an entry in the lookup-database.
 **/
static void checkDatabase() {
  int count;

  if (be_quiet == NO)
    printf("Checking Content Database\n");
  count = forEachEntryInDatabase(&indexContent);  
  if (be_quiet == NO) {
    if (be_verbose == NO)
      printf("\n");
    printf("==> Done checking %d entries in content database.\n", 
	   count);
  }
}

/**
 * Process a request to insert content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
static int checkInsert(GNUNET_TCP_SOCKET * sock,
		       TCP_Insert_Request * insertRequest) {
  CONTENT_Block *block;
  int i;
  int len;
  HashCode160 triple;
  ContentIndex entry;
  int res;  
  HexName hn;
  block = (CONTENT_Block*) xmalloc(sizeof(CONTENT_Block)*MAX_FSIZE,
    "checkInsert: block");

  if (ntohs(insertRequest->header.size) != sizeof(TCP_Insert_Request)) {
    sendTCPResult(sock, SYSERR);
    xfree(block, "checkInsert: block");
    return SYSERR;
  }
  hash2hex(&insertRequest->doubleHash,
	   &hn);
  if (be_verbose == YES)
    printf("* %s (ins)\n",
	   (char*)&hn);
  res = SYSERR;
  len = readContent(&insertRequest->doubleHash, 
		    sizeof(CONTENT_Block) * MAX_FSIZE,
		    &block[0]);
  for (i=0;i<len/sizeof(CONTENT_Block);i++) 
    if (0 == memcmp(&insertRequest->content,
		    &block[i],
		    sizeof(CONTENT_Block)) )      
      res = OK;
  /* it may still be ok, since padded content will
     not be identical if encrypted. To check for this,
     we'd have to decrypt all block[i]'s, the
     block that was passed in, and then compare the
     first "n" bytes. */
  if (res == SYSERR) {
    HashCode160 key;
    CONTENT_Block decryptMaster;
    CONTENT_Block decryptBlock;

    getLastKey(&key);
    if (OK == decryptContent(&insertRequest->content,
				 &key,
				 &decryptMaster))
      
    for (i=0;i<len/sizeof(CONTENT_Block);i++) {
      if (SYSERR == decryptContent(&block[i],
				   &key,
				   &decryptBlock))
	continue; /* skip */
      if (0 == memcmp(&decryptMaster,
		      &decryptBlock,
		      getLastLength()) )      
      res = OK;
    }
  }
  if (res == SYSERR) {
    HexName hn;

    hash2hex(&insertRequest->doubleHash,
	     &hn);
    if (be_quiet == NO)
      printf("Content %s missing in database. ",
	     (char*) &hn);
    if (do_fix == YES) {
      if (OK == writeContent(&insertRequest->doubleHash, 
			     &insertRequest->content)) {
	if (be_quiet == NO)
	  printf("Fixed.\n");
      } else {
	printf("WARNING: can not fix (database full?)\n");
      }
    } else 
      if (be_quiet == NO)
	printf("\n");
  }      

  /* also check/fix double-to-triple-hash lookup database */
  hash(&insertRequest->doubleHash,
       sizeof(HashCode160),
       &triple);
  if (NO == findEntry(&triple,
		      &entry)) {
    HexName hn;
    hash2hex(&insertRequest->doubleHash,
	     &hn);
    if (be_quiet == NO)
      printf("Content %s not indexed in triple-hash database. ",
	     (char*) &hn);
    if (do_fix == YES) {	
      indexContent(&insertRequest->doubleHash);
      if (be_quiet == NO)
	printf("Fixed.\n");
    } else
      if (be_quiet == NO)
	printf("\n");
  } else {
    if ( (ntohl(entry.fileNameIndex) != -1) ||
	 (ntohl(entry.fileOffset) != -1) ) {
      if (be_quiet == NO)
	printf("Bad value (II) stored in triple-hash database (could be ok) ");
      if (do_fix == YES) {
	indexContent(&insertRequest->doubleHash);
	if (be_quiet == NO)
	  printf("Fixed.\n");
      } else
	if (be_quiet == NO)
	  printf("\n");
    }
  }
  sendTCPResult(sock, OK);
  xfree(block, "checkInsert: block");

  return OK;
}

/**
 * Process a request to index content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
static int checkIndex(GNUNET_TCP_SOCKET * sock,
		      TCP_Indexing_Request * indexingRequest) {
  HashCode160 triple;
  ContentIndex res;
  HexName hn;

  hash2hex(&indexingRequest->contentIndex.doubleHash,
	   &hn);
  if (be_verbose == YES)
    printf("* %s (idx)\n",
	   (char*)&hn);
  hash(&indexingRequest->contentIndex.doubleHash,
       sizeof(HashCode160),
       &triple);
  if (ntohs(indexingRequest->header.size) != sizeof(TCP_Indexing_Request)) {
#if PRINT_TCP
    printf("TCP: WARNING: indexing request malformed!\n");
#endif
    sendTCPResult(sock, SYSERR);
    return SYSERR;
  }
#if PRINT_TCP 
  printf("TCP: received indexing request\n");
#endif
  /* check if everything is already in place, and if not and fixme
     is set, do the write-actions: */

  if (NO == findEntry(&triple,
		      &res)) {
    HexName hn;
    hash2hex(&indexingRequest->contentIndex.doubleHash,
	     &hn);
    if (be_quiet == NO)
      printf("Content %s not indexed in triple-hash database. ",
	     (char*) &hn);
    if (do_fix == YES) {	   
      addEntry(&indexingRequest->contentIndex);
      if (be_quiet == NO)
	printf("Fixed.\n");
    } else
      if (be_quiet == NO)
	printf("\n");
  } else {
    if (0 != memcmp(&res.doubleHash, 
		    &indexingRequest->contentIndex.doubleHash,
		    sizeof(HashCode160))) {
      if (be_quiet == NO)
	printf("Bad value (IR) stored in triple-hash database (could be ok) ");
      if (do_fix == YES) {
	addEntry(&indexingRequest->contentIndex);
	if (be_quiet == NO)
	  printf("Fixed.\n");
      } else
	if (be_quiet == NO)
	  printf("\n");
    }
  }
  sendTCPResult(sock, OK);
  return OK;
}

/**
 * Handle data available on the TCP socket descriptor;
 * check that the request is already fullfilled.
 **/
static void checkProcessor(int sockDescriptor){
  TCP_SOCKET_BUFFER buff;
  TYPED_TCP_HEADER * hdr;
  GNUNET_TCP_SOCKET sock;
  int i;

  /* register the socket */
  initGNUnetServerSocket(sockDescriptor,
			 &sock);
  hdr = (TYPED_TCP_HEADER*) &buff;
  while (1) {
    if (SYSERR == readFromSocket(&sock,
				 &buff) )
      break; /* connection closed */
    /* demultiplex */
    switch (ntohs(hdr->tcpType)) {
    case TCP_REQUEST_INSERT:
      i = checkInsert(&sock,
		      (TCP_Insert_Request*)hdr);
      break;
    case TCP_REQUEST_INDEX:
      i = checkIndex(&sock,
		     (TCP_Indexing_Request*) hdr);
      break;
    default:
      i = SYSERR;
      break;
    } /* end of switch */
    if (OK != i) {
      break;
    }
  }
  destroySocket(&sock);
}

/**
 * Check that the given file is properly indexed
 * (and fix if appropriate). Also return SYSERR
 * if the file is gone and should thus be removed
 * from the list.
 **/
int checkIndexedFile(FileName name,
		     int index) {
  int result;

  if (be_quiet == NO)
    printf("* %s\n",
	   name);
  result = insertFile(name,
		      "no description",
		      "no nameroot",
		      "no mimetype",
		      NULL,
		      0,
		      index,
		      LOCAL_INDEXED_CONTENT_PRIO, /* FIXME: should we use this?! */
		      NULL,
		      NULL);
  if (OK == result) 
    result = tcp_verifies;
  if (result == SYSERR) {
    if (be_quiet == NO)
      printf("Problem checking indexing of file %s ",
	     name);
    if (do_fix == YES) {
      if (be_quiet == NO)
	printf("Removing file from list.\n");
      return SYSERR; /* remove file, there was a problem */
    } else {
      if (be_quiet == NO)
	printf("\n");
      return OK;
    }
  }
  return OK;
}

/**
 * Check that all files that are listed in
 * the list of indexed files actually exist
 * and that they are properly indexed in the
 * lookup (triple->double hash) database.
 **/
static void checkIndexedFileList() {
  int count;

  if (be_quiet == NO)
    printf("Checking indexed files\n");
  count = forEachIndexedFile(&checkIndexedFile);
  if (be_quiet == NO)
    printf("==> Done with %d indexed files.\n",
	   count);
}

/**
 * Print a list of the options we offer.
 **/
static void printhelp() {
  printf("USAGE: gnunet-check [OPTIONS]\n");
  printf("-h, --help                    : print this page\n");
  printf("-c FILENAME, --config=FILENAME: load config file (defaults: %s)\n",
	 DEFAULT_CONFIG_FILE);
  printf("-v, --version                 : print the version number\n");
  printf("-n, --nofix                   : do not fix problems\n");
  printf("-a, --all                     : check all three databases\n");
  printf("-i, --index                   : only check the index database\n");
  printf("-d, --data                    : only check the content database\n");
  printf("-f, --files                   : only check the file-list\n");
  printf("-V, --verbose                 : be verbose\n");
  printf("-q, --quiet                   : be quiet\n");
  printf("Never run gnunet-check while gnunetd is running!\n");
}

/**
 * Perform option parsing from the command line. 
 **/
static int parseCommandLine(int argc, 
			    char * argv[],
			    char ** configFile,
			    char * check) {
  int shouldexit=0;
  int c;
  *check = 'n';
  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      { "config",  1, 0, 'c' },
      { "version", 0, 0, 'v' },
      { "help",    0, 0, 'h' },
      { "nofix",   0, 0, 'n' },
      { "index",   0, 0, 'i' },
      { "data",    0, 0, 'd' },
      { "files",   0, 0, 'f' },
      { "all",     0, 0, 'a' },
      { "verbose", 0, 0, 'V' },
      { "quiet",   0, 0, 'q' },
      { 0,0,0,0 }
    };
    
    LOOPPRINT("parseCommandLine"); 
    c = getopt_long(argc,
		    argv, 
		    "vhnc:dfaiVq", 
		    long_options, 
		    &option_index);
    
    if (c == -1) 
      break;  
    
    switch(c) {
    case 'q':
      be_quiet = YES;
      break;
    case 'a':
      *check = 'a';
      break;
    case 'i':
      *check = 'i';
      break;
    case 'f':
      *check = 'f';
      break;
    case 'V':
      be_verbose = YES;
      break;
    case 'd':
      *check = 'd';
      break;
    case 'n':
     do_fix = NO;
      break;
    case 'c': 
      *configFile = optarg; 
      printf("Configfile specified: %s.\n",*configFile);
      break;
    case 'v': 
      printf("gnunet-check %s (%s)\n",
	    GNUNET_VERSION, GNUNET_BRANCH);
      shouldexit=1;
      break;
    case 'h': 
      printhelp(); 
      shouldexit=1;
      break;
    default:
      printf("Unknown option %c. Aborting.\nUse --help to get a list of options.\n",
	    c);
      shouldexit=1;    
    } /* end of parsing commandline */
  }
  if (optind < argc) {
    printf("Invalid arguments: ");
    while (optind < argc)
      printf("%s ", argv[optind++]);
    printf("\nExiting.\n");
    return 1;
  }
  if ( (do_fix == NO) &&
       (be_quiet == NO) )
     printf("You selected verification only, will not fix problems!\n");
   return shouldexit;
}

/**
 * At the given position there is an entry that is
 * garbage, fix if we are allowed to.
 **/
static void deleteGarbage(CollisionLevel * level,
			  size_t pos) {
  int noll = 0;
  int i;

  if (do_fix == YES) {
    lseek(level->handle,
	  pos * sizeof(ContentIndex),
	  SEEK_SET);
    for (i=0;i<sizeof(ContentIndex)/sizeof(int);i++)
      write(level->handle, &noll, sizeof(int));
    if (be_quiet == NO)
      printf("Fixed\n");
  } else {
    if (be_quiet == NO)
      printf("\n");
  }
}

/**
 * Verify that the given entry from the 
 * triple-double-hash database is correct.
 **/
static int verifyEntryOk(ContentIndex * entry,
			 CollisionLevel * level,
			 size_t pos) {
  size_t expectPos;
  HashCode160 tripleHash;
  HexName hn;

  progressDot++;
  if ((progressDot & 255) == 0)
    if (be_quiet == NO) {
      printf(".");
      fflush(stdout);
    }
  hash(&entry->doubleHash,
       sizeof(HashCode160),
       &tripleHash);
  hash2hex(&entry->doubleHash,
	   &hn);
  if (be_verbose == YES)
    printf("* %s (idb)\n",
	   (char*)&hn);
  expectPos = ((int*)&tripleHash)[level->offset] & level->mask;
  if (pos - expectPos > LOOKUP_CORUN) {
    if (be_quiet == NO)
      printf("Entry %s in database is not at expected position (%d - %d > %d) ",
	     (char*)&hn, pos, expectPos, LOOKUP_CORUN);
    deleteGarbage(level, pos);
    return OK;
  }

  /* make sure that every item that occurs in the
     (triple-hash2double-hash) index is actually reachable and makes
     sense. */
  if (ntohl(entry->fileNameIndex) == -1) {
    int res;
    CONTENT_Block block;

    res = readContent(&entry->doubleHash,
		      sizeof(CONTENT_Block),
		      &block);
    if (-1 == res) {
      if (be_quiet == NO)
	printf("Indexed content %s not found in content database (%d). ",
	       (char*) &hn, res);
      deleteGarbage(level, pos);
      return OK;
    }
  } else {
    if ( (ntohl(entry->fileNameIndex) <= 0) || 
	 (ntohl(entry->fileNameIndex) > getSharedFileCount()) ) {
      if (be_quiet == NO)
	printf("Bad index into file list: %d ",
	       ntohl(entry->fileNameIndex));
      deleteGarbage(level, pos);
      return OK;      
    }
    if (SYSERR == checkHashMatch(ntohl(entry->fileNameIndex),
				 ntohl(entry->fileOffset),
				 &entry->doubleHash)) {
      printf("Indexed content at offset %d in file %s has changed. ",
	     ntohl(entry->fileOffset), 
	     indexed_files[ntohl(entry->fileNameIndex)-1]);
      deleteGarbage(level, pos);
      return OK;      
    }         
  }
  return OK;
}

/**
 * Verify that the triple-hash to double-hash mapping
 * database (lookup) is consistent.
 **/
static void checkIndexDatabase() {
  int count;

  if (be_quiet == NO)
    printf("Checking index database\n");
  count = forEachLookupEntry(&verifyEntryOk);
  if (be_quiet == NO) {
    if (be_verbose == NO)
      printf("\n");
    printf("==> Done checking %d entries in index (usage: %d%%).\n",
	   count, 
	   (int) (count*100.0/CACHE_HASH_TABLE_SIZE));
  }
}

int main(int argc, char * argv[]) {
  FileName configFile = DEFAULT_CONFIG_FILE;
  pthread_t tcpPseudoServer;
  char check;

  initLoggingFile(stderr,
		  LOG_MESSAGE,
		  NO);
  /* initialize subsystems */
  if (parseCommandLine(argc, argv, 
		       (char**)&configFile,
		       &check))
    return 0; /* parse error */
  if (check == 'n') {
    FAILURE("You must choose what to check (specify -d, -f, -i or -a).\n");
    return -1;
  }
  initConfiguration(configFile);

  PTHREAD_CREATE(&tcpPseudoServer,
		 (void * (*)()) &tcpListenMain, 
		 &checkProcessor,
     16 * 1024);
  sleep(1); /* ensure TCP server is started... */
  initContentDatabase(getContentDirectory());
  initCron();
  initLookup();
  initInsertfile("localhost");
  if ((check == 'a') || (check == 'd'))
    checkDatabase(); 
  if ((check == 'a') || (check == 'f'))
    checkIndexedFileList();
  if ((check == 'a') || (check == 'i'))
    checkIndexDatabase();

  doneInsertfile();
  doneConfiguration();
  return 0;
}


/* end of check_database.c */
